#pragma once
#include <easy2d/e2dmacros.h>
#include <easy2d/e2dcommon.h>
#include <easy2d/e2dtext.h>


// Base Classes

namespace easy2d
{


// Ϸ
class Game
{
public:
	// ʼϷ
	static bool init(
		const String& title = L"Easy2D",	/* ڱ */
		int width = 640,					/* ڿ */
		int height = 480,					/* ڸ߶ */
		const String& mutexName = L""		/* ̻ */
	);

	// Ϸ
	static void start(
		int fpsLimit = 0					/* FPS ƣ0Ϊ */
	);

	// ͣϷ
	static void pause();

	// Ϸ
	static void resume();

	// Ϸ
	static void quit();

	// ϷԴ
	static void destroy();

	// Ϸڲʱ
	static void reset();

	// ϷǷͣ
	static bool isPaused();

	// ȡϷ
	static String getName();
};


// ڿ
class Window
{
	friend class Game;

public:
	// ָʽ
	enum class Cursor : int
	{
		Normal,		/* Ĭָʽ */
		Hand,		/* ״ָ */
		No,			/* ָֹ */
		Wait,		/* ɳ©ָ */
		ArrowWait	/* ĬָСɳ© */
	};

public:
	// ޸ĴڴС
	static void setSize(
		int width,			/* ڿ */
		int height			/* ڸ߶ */
	);

	// ôڱ
	static void setTitle(
		const String& title	/* ڱ */
	);

	// ôͼ
	static void setIcon(
		int iconID
	);

	// ָʽ
	static void setCursor(
		Cursor cursor
	);

	// ȡڱ
	static String getTitle();

	// ȡڿ
	static float getWidth();

	// ȡڸ߶
	static float getHeight();

	// ȡڴС
	static Size getSize();

	// ȡھ
	static HWND getHWnd();

	// ǷӦ뷨
	static void setTypewritingEnable(
		bool enable
	);

	// ʾ
	static void info(
		const String& text,					/*  */
		const String& title = L"Infomation"	/* ڱ */
	);

	// 洰
	static void warning(
		const String& text,					/*  */
		const String& title = L"Warning"	/* ڱ */
	);

	// 󴰿
	static void error(
		const String& text,					/*  */
		const String& title = L"Error"		/* ڱ */
	);

private:
	// ʼ
	static bool __init(
		const String& title,
		int width,
		int height
	);

	// ô
	static void __uninit();

	// Ϣ
	static void __poll();

	// ָ
	static void __updateCursor();

	// Win32 Ϣص
	static LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
};


// ʱ
class Time
{
	friend class Game;

public:
	// ȡһ֡뵱ǰ֡ʱ룩
	static float getDeltaTime();

	// ȡһ֡뵱ǰ֡ʱ룩
	static unsigned int getDeltaTimeMilliseconds();

	// ȡϷʱ룩
	static float getTotalTime();

	// ȡϷʱ룩
	static unsigned int getTotalTimeMilliseconds();

private:
	// ʼʱ
	static void __init(int expectedFPS);

	// Ƿﵽʱ
	static bool __isReady();

	// µǰʱ
	static void __updateNow();

	// ʱϢ
	static void __updateLast();

	// ʱϢ
	static void __reset();

	// ߳
	static void __sleep();
};


class Listener;

// 
class Input
{
	friend class Game;

public:
	// ĳǷ
	static bool isDown(
		KeyCode::Value key
	);

	// ĳǷ񱻵
	static bool isPress(
		KeyCode::Value key
	);

	// ĳǷɿ
	static bool isRelease(
		KeyCode::Value key
	);

	// 갴Ƿ
	static bool isDown(
		MouseCode::Value code
	);

	// 갴Ƿ񱻵
	static bool isPress(
		MouseCode::Value code
	);

	// 갴Ƿɿ
	static bool isRelease(
		MouseCode::Value code
	);

	// Xֵ
	static float getMouseX();

	// Yֵ
	static float getMouseY();

	// ֵ
	static Point getMousePos();

	// X
	static float getMouseDeltaX();

	// Y
	static float getMouseDeltaY();

	// Zᣨ֣
	static float getMouseDeltaZ();

private:
	// ʼ DirectInput Լ豸
	static bool __init();

	// ˢϢ
	static void __update();

	// ж DirectInput
	static void __uninit();
};


// Ⱦ
class Renderer
{
	friend class Game;
	friend class Window;

public:
	// ȡɫ
	static Color getBackgroundColor();

	// ޸ıɫ
	static void setBackgroundColor(
		Color color
	);

	// ʾ FPS
	static void showFps(
		bool show = true
	);

	// ôֱͬԴɿ٣
	static void setVSync(
		bool enabled
	);

	// Ƿ˴ֱͬ
	static bool isVSyncEnabled();

	// ȡϵͳ DPI 
	static float getDpiScaleX();

	// ȡϵͳ DPI 
	static float getDpiScaleY();

	// ȡ ID2D1Factory 
	static ID2D1Factory * getID2D1Factory();

	// ȡ ID2D1HwndRenderTarget 
	static ID2D1HwndRenderTarget * getRenderTarget();

	// ȡ ID2D1SolidColorBrush 
	static ID2D1SolidColorBrush * getSolidColorBrush();

	// ȡ IWICImagingFactory 
	static IWICImagingFactory * getIWICImagingFactory();

	// ȡ IDWriteFactory 
	static IDWriteFactory * getIDWriteFactory();

	// ȡ Miter ʽ ID2D1StrokeStyle
	static ID2D1StrokeStyle * getMiterID2D1StrokeStyle();

	// ȡ Bevel ʽ ID2D1StrokeStyle
	static ID2D1StrokeStyle * getBevelID2D1StrokeStyle();

	// ȡ Round ʽ ID2D1StrokeStyle
	static ID2D1StrokeStyle * getRoundID2D1StrokeStyle();

	// Ⱦֲ
	static void DrawTextLayout(
		TextLayout* layout,
		const DrawingStyle& style,
		const Point& offset = Point(),
		ID2D1RenderTarget* rt = nullptr,
		ID2D1SolidColorBrush* brush = nullptr
	);

	// 豸ԴǷոؽ
	static bool isDeviceResourceRecreated();

private:
	// ȾϷ
	static void __render();

	// 豸޹Դ
	static bool __createDeviceIndependentResources();

	// 豸Դ
	static bool __createDeviceResources();

	// ɾ豸Դ
	static void __discardDeviceResources();

	// ɾȾԴ
	static void __discardResources();
};


// ־
class Logger
{
public:
	// ־¼
	static void enable();

	// ر־¼
	static void disable();

	// ־Ϣ
	static void messageln(String format, ...);

	// 
	static void warningln(String format, ...);

	// 
	static void errorln(String format, ...);

	// /رտ̨
	static void showConsole(bool show = true);
};


// װ
class GC
{
public:
	//  GC 
	static void trace(
		Object* pObject
	);

	// 
	static void clear();

	// Ƿ GC 
	static bool isInPool(Object* pObject);

	// GC ״̬
	static bool isClearing();

	// 
	template <typename Type>
	static inline void retain(Type*& p)
	{
		if (p != nullptr)
		{
			p->retain();
		}
	}

	// ͷŶ
	template <typename Type>
	static inline void release(Type*& p)
	{
		if (p != nullptr)
		{
			p->release();
			p = nullptr;
		}
	}
};


//
// GC macros
//

namespace __gc_helper
{
	class GCNewHelper
	{
	public:
		template <typename _Ty>
		inline _Ty* operator<< (_Ty* newObj) const
		{
			if (newObj)
			{
				newObj->autorelease();
			}
			return newObj;
		}

		// ʹ static  static inline 
		//  IntelliSense 
		static GCNewHelper instance;
	};
}

#ifndef gcnew
#	define gcnew __gc_helper::GCNewHelper::instance << new (std::nothrow)
#endif


//
// Log macros
//

#ifndef E2D_LOG
#	ifdef E2D_DEBUG
#		define E2D_LOG(FORMAT, ...) easy2d::Logger::messageln(FORMAT, __VA_ARGS__)
#	else
#		define E2D_LOG __noop
#	endif
#endif

#ifndef E2D_WARNING
#	define E2D_WARNING(FORMAT, ...) easy2d::Logger::warningln(FORMAT, __VA_ARGS__)
#endif

#ifndef E2D_ERROR
#	define E2D_ERROR(FORMAT, ...) easy2d::Logger::errorln(FORMAT, __VA_ARGS__)
#endif

#ifndef E2D_ERROR_IF_FAILED
#	define E2D_ERROR_IF_FAILED(HR, FORMAT, ...) do { if (FAILED(HR)) { E2D_ERROR(FORMAT, __VA_ARGS__); } } while (0)
#endif

}